"""SQL Session product


CREATE TABLE sessions (
  session varchar(40) UNIQUE NOT NULL,
  session_id number UNIQUE NOT NULL
   -- time?
);

CREATE TABLE session_data (
  session_id number NOT NULL,
  attr varchar(200) NOT NULL,
  value varchar(4000)
);


"""

__version__='$Revision: 1.4 $'[11:-2]

import Globals, App.Undo, socket, regex
from Products.ZSQLMethods.SQL import SQL
from Globals import PersistentMapping
from Globals import Persistent
from Globals import HTMLFile, MessageDialog
from string import join,strip,split,lower
from App.Management import Navigation, Tabs
from Acquisition import Implicit
import OFS.SimpleItem 
from OFS.Document import Document
from base64 import decodestring, encodestring
from urllib import quote, unquote
from ImageFile import ImageFile
import AccessControl.Role 
import string
from cPickle import loads, dumps

ListType=type([])


class SessionObj:
    """ A session object. Installed into the REQUEST object as 
	REQUEST.SESSION (like REQUEST.RESPONSE)
    """
    def __init__(self, parentSession, sessionName):
	#print "sessionobj init called"
	self.parentSession = parentSession
	self.sName = sessionName

    def __getitem__(self, name):
	#print "getitem called"
	r = self.parentSession.sqlGetValue(session=self.sName,
					    name=name)
	if len(r):
	    return loads(decodestring(sqlattr(r[0], 'value')))
	else:
	    return None  # hm. should it raise an error?

    def __setitem__(self, name, value):
	#print "setitem called"
	encvalue = encodestring(dumps(value))[:-1] # chop \n
	#print "encoded to", encvalue
	self.parentSession.sqlSetValue(session=self.sName,
				    name=name, value=encvalue)
	return None

    set = __setitem__

    def keys(self):
	rows = self.parentSession.sqlGetValue(session=self.sName)
	l = []
	for r in rows:
	   l.append(sqlattr(r, 'name'))
	return l

    def values(self):
	rows = self.parentSession.sqlGetValue(session=self.sName)
	l = []
	for r in rows:
	   l.append(loads(decodestring(sqlattr(r, 'value'))))
	return l

    def items(self):
	rows = self.parentSession.sqlGetValue(session=self.sName)
	l = []
	for r in rows:
	   name = sqlattr(r, 'name')
	   #print name
	   value = loads(decodestring(sqlattr(r, 'value')))
	   l.append(( name, value ))
	return l

    def has_key(self, name):
	#print "has_key called"
	r = self.parentSession.sqlGetValue(session=self.sName,
					    name=name)
	if len(r):
	    return 1
	else:
	    return 0

    def __str__(self):
	l = ["Session Variables:\n"]
	for n,v in self.items():
	    l.append("%s: %s"%(n,str(v)))
	return join(l, '\n')




class SQLSession(OFS.SimpleItem.Item, Persistent, Implicit, 
		      AccessControl.Role.RoleManager):
    """ An SQL Session. """


    meta_type='SQL Session'
    id       ='Session'
    title    ='SQL Session'
    icon     ='misc_/SQLSession/SQLSession_icon'

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Security',   'action':'manage_access'},
    )

    __ac_permissions__=(
    ('View management screens',
     ['manage','manage_menu','manage_main','manage_tabs' ]),
    ('Change SQLSessions',     ['manage_edit']),
    ('View', ['__call__','']),
    )

    manage_main=HTMLFile('main', globals())

    def __init__(self, conn_id, cookie_name):
        self.conn_id=conn_id
	self.cookie_name = cookie_name
        
        #self.docLogin =Document(_docLogin, __name__='docLogin')
        #self.docLogout=Document(_docLogout, __name__='docLogout')
        
        self.sqlSessionCreate=SQL('sqlSessionCreate', '', conn_id,
                              'session',
                              _sqlSessionCreate)
        self.sqlSessionLookup=SQL('sqlSessionLookup', '', conn_id,
                              'session',
                              _sqlSessionLookup)
        self.sqlGetValue=SQL('sqlGetValue', '', conn_id,
                              'session name=""',
                              _sqlGetValue)
        self.sqlSetValue=SQL('sqlSetValue', '', conn_id,
                             'session name value',
                              _sqlSetValue)

    def enumQueries(self):
        list=[]
        type=SQL.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
        return list

    def enumDocuments(self):
        list=[]
        type=Document.meta_type
        for ob in self.__dict__.values():
            if hasattr(ob, 'meta_type') and ob.meta_type==type:
                list.append(ob)
        return list

    def manage_edit(self, connection_id, cookie_name, REQUEST=None):
	"""Change properties"""
        if connection_id != self.conn_id:
            # Connection changed - update sql methods!
            self.conn_id=connection_id
            for ob in self.enumQueries():
                title=ob.title
                args=ob.arguments_src
                src=ob.src
                ob.manage_edit(title, self.conn_id, args, src)
	self.cookie_name = cookie_name
        if REQUEST:
            return MessageDialog(
                   title  ='SQLSesssion Changed',
                   message='SQLSesssion properties have been updated',
                   action =REQUEST['URL1']+'/manage_properties',
                   target ='manage_main')

    def __call__(self, REQUEST=None):
	""" magic bits here. """
	#print "called"
	if REQUEST is None:
	    try:
		REQUEST=self.REQUEST
	    except:
		raise "Session Error", "can't find REQUEST, even in self!"
	response = REQUEST['RESPONSE']
	if not REQUEST.cookies.has_key(self.cookie_name):
	    sessionName = self.newSession(REQUEST)
	else:
	    if REQUEST.has_key('_force_new_session'):
		sessionName = self.newSession(REQUEST)
	    else:
		sessionName = REQUEST.cookies[self.cookie_name]
		r = self.sqlSessionLookup(session=sessionName)
		if len(r) != 1:
		    sessionName = self.newSession(REQUEST)
		else:
		    pass # session is fine as-is.
	self.buildSessionObj(REQUEST, sessionName)
	return '<!-- session is "%s" -->'%sessionName

    def buildSessionObj(self, REQUEST, sessionName):
	s = SessionObj(self, sessionName)
	REQUEST['SESSION'] = s

    def newSession(self, REQUEST):
	response = REQUEST['RESPONSE']
	sessionName = self.makeSessionName(REQUEST)
	response.setCookie(self.cookie_name, sessionName, path='/')
	r = self.sqlSessionCreate(session=sessionName)
	return sessionName




    def makeSessionName(self, REQUEST):
	""" create a new unique string to name the session. feel free to
	    fiddle with this. """
	import md5, time, base64
	m = md5.new()
	m.update('happy happy happy boing plonk splut')
	m.update(str(time.time()))
	m.update(str(REQUEST))
	return base64.encodestring(m.digest())[:-3]



manage_addSQLSessionForm=HTMLFile('addSQLSession', globals())

def manage_addSQLSession(self, connection_id, cookie_name, REQUEST=None):
    """ """
    if hasattr(self.aq_base, 'Session'):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains a Session',
               action ='%s/manage_main' % REQUEST['PARENT_URL'])
    ob=SQLSession(connection_id, cookie_name)
    self._setObject('Session', ob)
    if REQUEST: return self.manage_main(self,REQUEST,update_menu=1)







# A getattr-like helper that hopefully can compensate for the
# various oddities in data returned by different stupidbases.


import Missing
mt=type(Missing.Value)

def typeconv(val):
    if type(val)==mt:
        return ''
    return val

def sqlattr(ob, attr):
    name=attr
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    attr=upper(attr)
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    attr=lower(attr)
    if hasattr(ob, attr):
        return typeconv(getattr(ob, attr))
    raise NameError, name

def absattr(attr):
    if callable(attr): return attr()
    return attr

def reqattr(request, attr):
    try:    return request[attr]
    except: return None

_sqlSessionCreate = """<dtml-comment>

This SQL creates a new session.

</dtml-comment>
INSERT INTO Sessions 
( session )
values
( <dtml-sqlvar session type="string"> )

"""

_sqlSessionLookup = """<dtml-comment>
This query looks up a particular session. It 
should return one or more rows if the session 
is active.
</dtml-comment>

SELECT * FROM Sessions
WHERE 
<dtml-sqltest session type=string>

"""

_sqlGetValue = """<dtml-comment>
This query looks up some or all data belonging 
to a session.
</dtml-comment>

SELECT * FROM Session_data
WHERE 
<dtml-sqltest session type=string>
<dtml-if name>
AND <dtml-sqltest name type=string>
</dtml-if name>

"""

_sqlSetValue = """<dtml-comment>
This query sets or updates a data item belonging 
to a session.  Updates are bad, so we delete/insert.
</dtml-comment>

DELETE FROM Session_data
WHERE 
<dtml-sqltest session type=string>
AND 
<dtml-sqltest name type=string>

<dtml-var sql_delimiter>

INSERT INTO Session_data
( session, name, value ) 
VALUES
( <dtml-sqlvar session type="string">,
  <dtml-sqlvar name type="string">,
  <dtml-sqlvar value type="string"> )

"""
